(require 'cl)

(require 'cleite-constants)
(require 'cleite-vars)
(require 'cleite-face)
(require 'cleite-feeds)
(require 'cleite-items)
(require 'cleite-article)
(require 'cleite-dialog)
(require 'cleite-w3m)
(require 'cleite-system)
(require 'srpc-active-node)
(require 'timer)

(defvar cleite:current-node nil
  "Current client node which communicates with the cleite daemon")

(defvar cleite:current-password "")

(defvar cleite:auto-refresh-timer nil)

(defun cleite:set-current-password (new-password)
  (setq cleite:current-password new-password))

(defun cleite:node (&optional always-check-account)
  "Return the current client node which communicates with the cleite daemon.

If ALWAYS-CHECK-ACCOUNT is t, check the validity of account even if
the node has already been created and checked before."
  
  ;; 現在のノードが nil であれば新たに作成すると共にチェック。
  ;; nil でなくても明示的に指示されてゐればチェック。
  (if (null cleite:current-node)
	  (progn
		(setq cleite:current-node (srpc:active-node "cleite"
													:host cleite:daemon-host
													:port cleite:daemon-port))
		(cleite:check-account))

	(if always-check-account
		(cleite:check-account)))

  cleite:current-node)


(defun cleite:check-account ()
  (let ((node cleite:current-node))
	(condition-case nil
		(srpc:call node 'account:check
				   `((user-id . ,cleite:user-id         )
					 (passwd  . ,cleite:current-password)))
	  (srpc:failure
	   (setq cleite:current-password
			 (read-passwd
			  (format "Enter the password for user `%s': " cleite:user-id)))
	   (cleite:check-account)))))


(defun cleite:make-time-float (list)
  (+ (* 65536.0 (pop list)) (pop list)
     (/ (or (pop list) 0) 1e6)))


(defun cleite:call (method &optional args-alist)
  (let ((begin (cleite:make-time-float (current-time))))

	(prog1
		(srpc:call (cleite:node t)
				   method
				   (append `((auth . ((user-id . ,cleite:user-id         )
									  (passwd  . ,cleite:current-password))))
						   args-alist))

	  (when cleite:measure-srpc-call-time
		(message "Time elapsed: %.4f  - Method: %s"
				 (- (cleite:make-time-float (current-time))
					begin)
				 method)))))


(defun cleite:fast-call (method &optional args-alist)
  (srpc:call (cleite:node nil)
			 method
			 (append `((auth . ((user-id . ,cleite:user-id         )
								(passwd  . ,cleite:current-password))))
					 args-alist)))


(defun cleite:faced-string (str face)
  (let ((faced (apply 'string (string-to-list str))))
	
	(add-text-properties 0 (length faced)
						 (list 'face face)
						 faced)
	faced))


(defun cleite:read-char (&optional prompt)
  (let ((cursor-in-echo-area t  )
		(ch                  nil))
	
	(if prompt
		(message "%s" prompt))

	(setq ch (read-char))

	(if (and prompt
			 (char-valid-p ch))
		(message "%s%c" prompt ch))

	ch))

(defun cleite:read-char-with-retry (prompt candidates)
  (let ((ch (cleite:read-char prompt)))
	  
	(cond ((memq ch candidates)
		   ch)

		  ((eq ch 12)
		   (recenter)
		   (cleite:read-char-with-retry prompt candidates))

		  (t
		   (ding)
		   (cleite:read-char-with-retry prompt candidates)))))


(defmacro cleite:with-temp-window (buffer-name &rest rest)
  (declare (indent 1))
  `(save-window-excursion
	 (delete-other-windows)

	 (let ((buffer (generate-new-buffer ,buffer-name)))
	   (unwind-protect
		   ;; このバッファを表示する爲に畫面を上下に分割する。
		   (let ((window (split-window-vertically)))
			 (select-window    window)
			 (switch-to-buffer buffer)
			 (erase-buffer)

			 ,@rest)
		 (kill-buffer buffer)))))


(defun cleite:choose-element (prompt candidates)
  "選択肢の一覽を一時的なバッファに表示して選ばせる。

PROMPT は minibuffer に表示する文字列。
CANDIDATES は次のやうなリスト。

  ((\"選択肢１ -- １行目\"
    \"選択肢１ -- ２行目\"
    ...)

   (\"選択肢２ -- １行目\"
    ...))"
  (cleite:with-temp-window "*candidates*"
	(let ((n-cands (length candidates)))
	  
	  ;; バッファに選択肢一覽を書き込む
	  (loop for i from 1 to n-cands do
			(let ((cand (nth (1- i) candidates))
				  (label-str (format "[%d]  " i)))

			  (insert "\n")
			  (insert label-str)
			  (insert (car cand))
			  (insert "\n")
			  (dolist (line (cdr cand))
				(insert (make-string (length label-str) ?\ ))
				(insert line)
				(insert "\n"))))
	  
	  (setq buffer-read-only t)

	  ;; 項目の數が９個以下なら cleite:read-char-with-retry を使
	  ;; へるが、１０個以上あるなら使へないので補完入力を行ふ。
	  (if (> 10 n-cands)
		  (let* ((choice-ch (cleite:read-char-with-retry
							 (format "%s [%d-%d] " prompt 1 n-cands)
							 (loop for i from 1 to n-cands collect
								   (+ ?0 i))))
				 (choice    (- choice-ch ?0)))
			(1- choice))

		(labels ((choose ()
				     (let ((str (completing-read prompt
								    (loop for i from 1 to n-cands collect
										  (format "%d" i))
									nil  ;; predicate
									1))) ;; require-match
					   (if (equal str "")
						   (choose)
						 (1- (string-to-number str))))))
		  (choose))))))
	  

;;;###autoload
(defun cleite ()
  (interactive)
  (cleite:quit t)
  (cleite:feeds)
  (cleite:start-timer-if-needed))


(defun cleite:quit (&optional do-not-forget-password)
  (interactive "P")
  (cleite:stop-timer-if-any)
  (cleite:article:quit t)
  (cleite:items:quit t)
  (cleite:feeds:kill-buffer)
  (when cleite:current-node
    (srpc:disconnect cleite:current-node))
  (unless do-not-forget-password
    (setq cleite:current-password "")))


(defun cleite:start-timer-if-needed ()
  (when (and cleite:auto-refresh-interval
             (> cleite:auto-refresh-interval 0))
    (setq cleite:auto-refresh-timer
          (run-with-timer (* 60 cleite:auto-refresh-interval) ;; sec
                          (* 60 cleite:auto-refresh-interval) ;; repeat
                          (lambda ()
                            (condition-case exn
                                (progn
                                  (cleite:feeds:refresh)
                                  (cleite:items:refresh)
                                  )
                              (error (message "auto-refresh failed: %S" exn))))))))


(defun cleite:stop-timer-if-any ()
  (when cleite:auto-refresh-timer
    (cancel-timer cleite:auto-refresh-timer)
    (setq cleite:auto-refresh-timer nil)))


(defun cleite:version ()
  (interactive)
  (message "ElCleite - Elisp client for the server-based RSS aggregator Cleite - %s"
		   cleite:constants:version))


(defun cleite:root-p (&optional user-id)
  (let ((account (cleite:call 'account:get
                              `((user-id . ,(if user-id
                                                user-id
                                              cleite:user-id))))))
    (eq 'true (cdr (assq 'is-root account)))))
        

(defun cleite:assert-i-am-root ()
  (if (cleite:root-p)
      t
    (error "Operation not permitted")))


(provide 'cleite)